"""Class definition for ChatBot."""

import contextvars

import panel as pn
import tiktoken
from langchain.callbacks.base import BaseCallbackManager
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from langchain.chat_models import ChatOpenAI
from langchain.schema import AIMessage, HumanMessage, SystemMessage
from loguru import logger

from llamabot.config import default_language_model
from llamabot.panel_utils import PanelMarkdownCallbackHandler
from llamabot.recorder import autorecord

prompt_recorder_var = contextvars.ContextVar("prompt_recorder")

# Reference: https://platform.openai.com/docs/models
model_chat_token_budgets = {
    "gpt-3.5-turbo-0301": 4_097,
    "gpt-3.5-turbo-0613": 4_097,
    "gpt-3.5-turbo-16k-0613": 16_385,
    "gpt-3.5-turbo-16k": 16_385,
    "gpt-3.5-turbo-instruct-0914": 4_097,
    "gpt-3.5-turbo-instruct": 4_097,
    "gpt-3.5-turbo": 4_097,
    "gpt-4-0314": 8_192,
    "gpt-4-0613": 8_192,
    "gpt-4-32k": 32_768,
    "gpt-4-32k-0613": 32_768,
    "gpt-4": 8_192,
}


class ChatBot:
    """Chat Bot that is primed with a system prompt, accepts a human message.

    Automatic chat memory management happens.

    h/t Andrew Giessel/GPT4 for the idea.
    """

    def __init__(
        self,
        system_prompt,
        temperature=0.0,
        model_name=default_language_model(),
        logging=False,
        response_budget=2_000,
    ):
        """Initialize the ChatBot.

        :param system_prompt: The system prompt to use.
        :param temperature: The model temperature to use.
            See https://platform.openai.com/docs/api-reference/completions/create#completions/create-temperature
            for more information.
        :param model_name: The name of the OpenAI model to use.
        :param logging: Whether to log the chat history.
        :param response_budget: The number of tokens to budget for the response.
        """
        self.model = ChatOpenAI(
            model_name=model_name,
            temperature=temperature,
            streaming=True,
            verbose=True,
            callback_manager=BaseCallbackManager(
                handlers=[StreamingStdOutCallbackHandler()]
            ),
        )
        self.chat_history = [
            SystemMessage(content="Always return Markdown-compatible text."),
            SystemMessage(content=system_prompt),
        ]
        self.logging = logging
        self.model_name = model_name
        self.response_budget = response_budget

    def __call__(self, human_message) -> AIMessage:
        """Call the ChatBot.

        :param human_message: The human message to use.
        :return: The response to the human message, primed by the system prompt.
        """
        self.chat_history.append(HumanMessage(content=human_message))

        # Get out the last 6000 tokens of chat history.
        faux_chat_history = []
        enc = tiktoken.encoding_for_model("gpt-4")
        token_budget = model_chat_token_budgets[self.model_name] - self.response_budget

        # Put the system message into the faux chat history.
        system_messages = []
        for message in self.chat_history:
            if isinstance(message, SystemMessage):
                system_messages.append(message)
                tokens = enc.encode(message.content)
                token_budget -= len(tokens)

        faux_chat_history = []
        for message in self.chat_history[::-1]:
            if token_budget > 0:
                if self.logging:
                    logger.info(f"Token budget: {token_budget}")
                faux_chat_history.append(message)
                tokens = enc.encode(message.content)
                token_budget -= len(tokens)

        messages = system_messages + faux_chat_history[::-1]
        if self.logging:
            logger.info(messages)

        response = self.model(messages)
        self.chat_history.append(response)
        autorecord(human_message, response.content)

        return response

    def __repr__(self):
        """Return a string representation of the ChatBot.

        :return: A string representation of the ChatBot.
        """
        representation = ""

        for message in self.chat_history:
            if isinstance(message, SystemMessage):
                prefix = "[System]\n"
            elif isinstance(message, HumanMessage):
                prefix = "[Human]\n"
            elif isinstance(message, AIMessage):
                prefix = "[AI]\n"

            representation += f"{prefix}{message.content}" + "\n\n"
        return representation

    def panel(
        self,
        show: bool = True,
        site="ChatBot",
        title="ChatBot",
        width=768,
    ):
        """Create a Panel app that wraps a LlamaBot.

        :param show: Whether to show the app.
            If False, we return the Panel app directly.
            If True, we call `.show()` on the app.
        :param site: The name of the site.
        :param title: The title of the app.
        :param width: The width of the app in pixels.
        :return: The Panel app, either showed or directly.
        """

        text_input = pn.widgets.TextAreaInput(placeholder="Start chatting...")
        chat_history = pn.Column(*[])
        send_button = pn.widgets.Button(name="Send", button_type="primary")

        widget_width = width - 150

        def b(event):
            """Button click handler.

            :param event: The button click event.
            """
            chat_messages = []
            for message in self.chat_history:
                if isinstance(message, SystemMessage):
                    pass
                elif isinstance(message, HumanMessage):
                    chat_markdown = pn.pane.Markdown(
                        f"Human: {message.content}", width=widget_width
                    )
                    chat_messages.append(chat_markdown)
                elif isinstance(message, AIMessage):
                    chat_markdown = pn.pane.Markdown(
                        f"Bot: {message.content}", width=widget_width
                    )
                    chat_messages.append(chat_markdown)

            chat_messages.append(
                pn.pane.Markdown(f"Human: {text_input.value}", width=widget_width)
            )
            bot_reply = pn.pane.Markdown("Bot: ", width=widget_width)
            chat_messages.append(bot_reply)
            chat_history.objects = chat_messages
            markdown_handler = PanelMarkdownCallbackHandler(bot_reply)
            self.model.callback_manager.set_handler(markdown_handler)
            self(text_input.value)
            text_input.value = ""

        send_button.on_click(b)
        input_pane = pn.Row(text_input, send_button)
        output_pane = pn.Column(chat_history, scroll=False, height=500)

        main = pn.Row(input_pane, output_pane)
        app = pn.template.FastListTemplate(
            site=site,
            title=title,
            main=main,
            main_max_width=f"{width}px",
        )
        if show:
            return app.show()
        return app



"""Git subcommand for LlamaBot CLI."""
from pathlib import Path
import os
from tempfile import NamedTemporaryFile

import git
from pyprojroot import here
from sh import pre_commit
from rich.progress import Progress, SpinnerColumn, TextColumn
from typer import Typer, echo

from llamabot import SimpleBot
from llamabot.cli.utils import get_valid_input
from llamabot.code_manipulation import get_git_diff
from llamabot.prompt_library.git import (
    commitbot,
    write_commit_message,
    compose_release_notes,
)

gitapp = Typer()


@gitapp.command()
def commit(autocommit: bool = True):
    """Commit staged changes.

    :param autocommit: Whether to automatically commit the changes.
    """
    # Run pre-commit hooks first so that we don't waste tokens if hooks fail.
    repo = git.Repo(search_parent_directories=True)
    bot = commitbot()
    progress = Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    )

    while True:
        diff = get_git_diff()
        if not diff:
            print(
                "I don't see any staged changes to commit! "
                "Please stage some files before running me again."
            )
            return
        pre_commit("run")

        message = bot(write_commit_message(diff))
        print("\n\n")
        user_response = get_valid_input(
            "Do you accept this commit message?", ("y", "n", "m")
        )
        if user_response == "y":
            break
        elif user_response == "m":
            # Provide option to edit the commit message
            with NamedTemporaryFile(mode="w+", delete=False) as temp_file:
                temp_file.write(message.content)
                temp_file.flush()
                editor = os.getenv(
                    "EDITOR", "nano"
                )  # Use the system's default text editor
                os.system(f"{editor} {temp_file.name}")
                temp_file.seek(0)
                edited_message = temp_file.read().strip()
                message.content = edited_message
                break

    if autocommit:
        with progress:
            progress.add_task("Committing changes", total=None)
            repo.index.commit(message.content)
            progress.add_task("Pushing changes", total=None)
            origin = repo.remote(name="origin")
            origin.push()


@gitapp.command()
def install_commit_message_hook():
    """Install a commit message hook that runs the commit message through the bot.

    :raises RuntimeError: If the current directory is not a git repository root.
    """
    # Check that we are in a repository's root. There should be a ".git" folder.
    # Use pathlib to verify.
    if not Path(".git").exists():
        raise RuntimeError(
            "You must be in a git repository root folder to use this command. "
            "Please `cd` into your git repo's root folder and try again, "
            "or use `git init` to create a new repository (if you haven't already)."
        )

    with open(".git/hooks/prepare-commit-msg", "w+") as f:
        contents = """#!/bin/sh
llamabot git compose-commit > .git/COMMIT_EDITMSG
"""
        f.write(contents)
    os.chmod(".git/hooks/prepare-commit-msg", 0o755)
    echo("Commit message hook successfully installed! 🎉")


@gitapp.command()
def compose_commit():
    """Autowrite commit message based on the diff."""
    try:
        diff = get_git_diff()
        bot = commitbot()
        bot(write_commit_message(diff))
    except Exception as e:
        echo(f"Error encountered: {e}", err=True)
        echo("Please write your own commit message.", err=True)


@gitapp.command()
def write_release_notes(release_notes_dir: Path = Path("./docs/releases")):
    """Write release notes for the latest two tags to the release notes directory.

    :param release_notes_dir: The directory to write the release notes to.
        Defaults to "./docs/releases".
    """
    repo = git.Repo(here())
    tags = sorted(repo.tags, key=lambda t: t.commit.committed_datetime)
    tag1, tag2 = tags[-2], tags[-1]
    log_info = repo.git.log(f"{tag1.commit.hexsha}..{tag2.commit.hexsha}")

    bot = SimpleBot(
        "You are an expert software developer "
        "who knows how to write excellent release notes based on git commit logs."
    )
    response = bot(
        "Please change the following git remote URL into its most probable HTTPS URL: "
        "{repo.remotes.origin.url}. Return only the URL and nothing else."
    )
    repo_url = response.content

    notes = bot(compose_release_notes(log_info, repo_url))

    # Create release_notes_dir if it doesn't exist:
    release_notes_dir.mkdir(parents=True, exist_ok=True)

    # Write release notes to the file:
    with open(release_notes_dir / f"{tag2.name}.md", "w+") as f:
        f.write(notes.content)



"""Document chat bots."""
from pathlib import Path

import typer

from llamabot import QueryBot

from .utils import exit_if_asked, uniform_prompt

app = typer.Typer()


@app.command()
def chat(
    doc_path: Path = typer.Argument(
        "", help="Path to the document you wish to chat with."
    )
):
    """Chat with your document.

    :param doc_path: Path to the document you wish to chat with.
    """
    bot = QueryBot(
        system_message="You are a bot that can answer questions about a document provided to you.",
        doc_path=doc_path,
    )
    typer.echo(
        "I've embedded your document. Ask me anything! Otherwise, type 'exit' or 'quit' at anytime to exit."
    )

    while True:
        query = uniform_prompt()
        exit_if_asked(query)
        bot(query)
        typer.echo("\n\n")



"""
This module provides a Panel Markdown callback handler for streaming.

Classes:
    - PanelMarkdownCallbackHandler: Callback handler for streaming. Only works with LLMs that support streaming.

Exceptions:
    - None

Functions:
    - None
"""
from typing import Any, Dict, List, Union

import panel as pn
from langchain.callbacks.base import BaseCallbackHandler
from langchain.schema import AgentAction, AgentFinish, LLMResult


class PanelMarkdownCallbackHandler(BaseCallbackHandler):
    """Callback handler for streaming. Only works with LLMs that support streaming."""

    def __init__(self, markdown_object: pn.pane.Markdown):
        self.md = markdown_object

    def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> None:
        """Run when LLM starts running.

        # noqa: DAR101
        """

    def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """Run on new LLM token. Only available when streaming is enabled.

        # noqa: DAR101
        """
        self.md.object += f"{token}"

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        """Run when LLM ends running.

        # noqa: DAR101
        """

    def on_llm_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        """Run when LLM errors.

        # noqa: DAR101
        #"""

    def on_chain_start(
        self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> None:
        """Run when chain starts running.

        # noqa: DAR101
        """

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> None:
        """Run when chain ends running.

        # noqa: DAR101
        #"""

    def on_chain_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        """Run when chain errors.

        # noqa: DAR101
        """

    def on_tool_start(
        self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> None:
        """Run when tool starts running.

        # noqa: DAR101
        """

    def on_agent_action(self, action: AgentAction, **kwargs: Any) -> Any:
        """Run on agent action.

        # noqa: DAR101
        """
        pass

    def on_tool_end(self, output: str, **kwargs: Any) -> None:
        """Run when tool ends running.

        # noqa: DAR101
        """

    def on_tool_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        """Run when tool errors.

        # noqa: DAR101
        """

    def on_text(self, text: str, **kwargs: Any) -> None:
        """Run on arbitrary text.

        # noqa: DAR101
        """

    def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> None:
        """Run on agent end.

        # noqa: DAR101
        """



"""Llamabot Zotero CLI."""
from datetime import date
from pathlib import Path

import typer
from caseconverter import snakecase
from dotenv import load_dotenv
from prompt_toolkit import prompt
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn

from llamabot import QueryBot
from llamabot.recorder import PromptRecorder
from llamabot.zotero.library import ZoteroItem, ZoteroLibrary
from llamabot.zotero.completer import PaperTitleCompleter
from .utils import configure_environment_variable, exit_if_asked, uniform_prompt

load_dotenv()

app = typer.Typer()

ZOTERO_JSON_DIR = Path.home() / ".llamabot/zotero/zotero_index/"
ZOTERO_JSON_DIR.mkdir(parents=True, exist_ok=True)
progress = Progress(
    SpinnerColumn(),
    TextColumn("[progress.description]{task.description}"),
    transient=False,
)


console = Console()


@app.command()
def configure(
    library_id: str = typer.Option(..., prompt=True),
    api_key: str = typer.Option(..., prompt=True),
    library_type: str = typer.Option(default="user", prompt=True),
):
    """Configure Llamabot Zotero CLI environment variables.

    :param library_id: Zotero library ID
    :param api_key: Zotero API key
    :param library_type: Zotero library type
    """
    configure_environment_variable("ZOTERO_LIBRARY_ID", library_id)
    configure_environment_variable("ZOTERO_API_KEY", api_key)
    configure_environment_variable("ZOTERO_LIBRARY_TYPE", library_type)


@app.command()
def chat(
    query: str = typer.Argument("", help="The thing you want to chat about."),
    sync: bool = typer.Option(
        True, help="Whether or not to synchronize the Zotero library."
    ),
):
    """Chat with a paper.

    :param query: A paper to search for, whether by title, author, or other metadata.
    :param sync: Whether or not to synchronize the Zotero library.
    """
    typer.echo("Llamabot Zotero Chatbot initializing...")
    typer.echo("Use Ctrl+C to exit anytime.")

    if sync:
        library = ZoteroLibrary()
        library.to_json(ZOTERO_JSON_DIR)
    else:
        library = ZoteroLibrary(json_dir=ZOTERO_JSON_DIR)

    completer = PaperTitleCompleter(library.key_title_map().values())
    while True:
        user_choice = prompt(
            "Please choose an paper: ",
            completer=completer,
            complete_while_typing=True,
        )
        if user_choice in library.key_title_map().values():
            break
    typer.echo(f"Awesome! You have chosen the paper: {user_choice}")

    paper_key = library.key_title_map(inverse=True)[user_choice.strip(" ")]

    # Retrieve paper from library
    with progress:
        task = progress.add_task("Downloading paper...")
        entry: ZoteroItem = library[paper_key]
        fpath = entry.download_pdf(Path("/tmp"))
        progress.remove_task(task)

    typer.echo(f"Downloaded paper to {fpath}")

    with progress:
        task = progress.add_task("Embedding paper and initializing bot...")
        docbot = QueryBot(
            "You are an expert in answering questions about a paper.",
            doc_paths=[fpath],
        )
        progress.remove_task(task)

    # From this point onwards, we need to record the chat.
    pr = PromptRecorder()
    date_str = date.today().strftime("%Y%m%d")
    snaked_user_choice = f"{snakecase(user_choice)}"
    save_path = Path(f"{date_str}_{snaked_user_choice}.md")
    with pr:
        typer.echo("\n\n")
        typer.echo("Here is a summary of the paper for you to get going:")
        docbot("What is the summary of the paper?")
        typer.echo("\n\n")
        pr.save(save_path)

    while True:
        with pr:
            query = uniform_prompt()
            exit_if_asked(query)
            docbot(query)
            typer.echo("\n\n")

            # Want to append YYYYMMDD before filename.
            pr.save(save_path)



"""
A Python bot for generating module-level and function docstrings, and code generation.

This module provides a command-line interface for generating module-level and function
docstrings, as well as generating code based on a given description. It exports the
following commands:

- module_docstrings: Generate module-level docstrings for a given module file.
- function_docstrings: Generate function docstrings for a specific function in a module file.
- code_generator: Generate code based on a given description.

Classes:
    None

Exceptions:
    None

Functions:
    module_docstrings(module_fpath: Path)
    function_docstrings(module_fpath: Path, function_name: str, style: str = "sphinx")
    code_generator(request: str)
"""

from pathlib import Path

import pyperclip
from typer import Typer

from llamabot.code_manipulation import (
    get_dependencies,
    get_object_source_code,
    show_directory_tree,
)
from llamabot.file_finder import read_file
from llamabot.prompt_library.python import (
    codebot,
    docstring,
    ghostwriter,
    module_doc,
    tests,
)

from .utils import get_valid_input

app = Typer()


@app.command()
def module_docstrings(module_fpath: Path, dirtree_context_path: Path = None):
    """Generate module-level docstrings.

    :param module_fpath: Path to the module to generate docstrings for.
    :param dirtree_context_path: Path to the directory to use as the context for the
        directory tree.
        Defaults to the parent directory of the module file.
    """
    cb = codebot()
    while True:
        source_code = read_file(module_fpath)
        tree_structure = show_directory_tree(
            dirtree_context_path if dirtree_context_path else module_fpath.parent
        )
        module_docstring = cb(module_doc(source_code, module_fpath, tree_structure))
        print("\n\n")
        user_response = get_valid_input("Do you accept this docstring? (y/n) ")
        if user_response == "y":
            break
    pyperclip.copy(module_docstring.content)
    print("Copied to clipboard!")


@app.command()
def generate_docstrings(module_fpath: Path, object_name: str, style: str = "sphinx"):
    """Generate function docstrings.

    :param module_fpath: Path to the module to generate docstrings for.
    :param object_name: Name of the object to generate docstrings for.
    :param style: Style of docstring to generate.
    """
    cb = codebot()
    while True:
        # source_code = read_file(module_fpath)
        object_source = get_object_source_code(module_fpath, object_name)
        object_docstring = cb(docstring(object_source, style=style))
        print("\n\n")
        user_response = get_valid_input("Do you accept this docstring? (y/n) ")
        if user_response == "y":
            break

    pyperclip.copy(object_docstring.content)
    print("Copied to clipboard!")


@app.command()
def code_generator(request: str):
    """
    Generate code.

    .. code-block:: python

        code_generator("Create a function that adds two numbers")

    :param request: A description of what the code should do.
    """
    cb = codebot()
    while True:
        code = cb(ghostwriter(request, "Python"))
        print("\n\n")
        user_response = get_valid_input("Do you accept this code? (y/n) ")
        if user_response == "y":
            break
    pyperclip.copy(code.content)
    print("Copied to clipboard!")


@app.command()
def test_writer(module_fpath: str, object_name: str):
    """Write tests for a given object.

    :param module_fpath: Path to the module to generate tests for.
    :param object_name: Name of the object to generate tests for.
    """
    cb = codebot()
    while True:
        file_source = read_file(module_fpath)
        function_source = get_object_source_code(module_fpath, object_name)

        # Get other source files.
        other_source_files = ""
        for file in get_dependencies(module_fpath, object_name):
            if file != module_fpath:
                other_source_files += f"# {file}"
                other_source_files += read_file(file) + "\n\n"

        test_code = cb(tests(function_source, file_source, other_source_files))
        print("\n\n")
        user_response = get_valid_input("Do you accept this test? (y/n) ")
        if user_response == "y":
            break
    pyperclip.copy(test_code.content)
    print("Copied to clipboard!")



"""Class definition for QueryBot."""
import contextvars
import hashlib
import os
from pathlib import Path
from typing import List, Union

from langchain.callbacks.base import BaseCallbackManager
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from langchain.chat_models import ChatOpenAI
from langchain.schema import AIMessage, HumanMessage, SystemMessage
from llama_index import (
    Document,
    GPTVectorStoreIndex,
    LLMPredictor,
    ServiceContext,
    StorageContext,
    load_index_from_storage,
)
from llama_index.indices.vector_store.retrievers.retriever import VectorIndexRetriever
from llama_index.node_parser import SimpleNodeParser
from llama_index.storage.docstore import SimpleDocumentStore
from llama_index.storage.index_store import SimpleIndexStore
from llama_index.vector_stores import SimpleVectorStore

from llamabot.config import default_language_model
from llamabot.doc_processor import magic_load_doc, split_document
from llamabot.recorder import autorecord

DEFAULT_SIMILARITY_TOP_KS = {
    "gpt-3.5-turbo": 2,
    "gpt-3.5-turbo-0301": 2,
    "gpt-3.5-turbo-0613": 2,
    "gpt-3.5-turbo-16k": 5,
    "gpt-3.5-turbo-16k-0613": 5,
    "gpt-4": 3,
    "gpt-4-0314": 3,
    "gpt-4-0613": 3,
    "gpt-4-32k": 10,
}


SIMILARITY_TOP_K = DEFAULT_SIMILARITY_TOP_KS.get(os.getenv("OPENAI_DEFAULT_MODEL"), 3)


CACHE_DIR = Path.home() / ".llamabot" / "cache"
prompt_recorder_var = contextvars.ContextVar("prompt_recorder")


class QueryBot:
    """QueryBot is a bot that lets us use GPT4 to query documents."""

    def __init__(
        self,
        system_message: str,
        model_name: str = default_language_model(),
        temperature: float = 0.0,
        doc_paths: List[str] | List[Path] | str | Path = None,
        saved_index_path: str | Path = None,
        chunk_size: int = 2000,
        chunk_overlap: int = 0,
        stream: bool = True,
        use_cache: bool = True,
    ):
        """Initialize QueryBot.

        Pass in either the doc_path or saved_index_path to initialize the QueryBot.

        NOTE: QueryBot is not designed to have memory!

        The default text splitter is the TokenTextSplitter from LangChain.
        The default index that we use is the GPTVectorStoreIndex from LlamaIndex.
        We also default to using GPT4 with temperature 0.0.

        :param system_message: The system message to send to the chatbot.
        :param model_name: The name of the OpenAI model to use.
        :param temperature: The model temperature to use.
            See https://platform.openai.com/docs/api-reference/completions/create#completions/create-temperature
            for more information.
        :param doc_paths: A path to a document,
            or a list of paths to multiple documents,
            to use for the chatbot.
        :param saved_index_path: The path to the saved index to use for the chatbot.
        :param chunk_size: The chunk size to use for the LlamaIndex TokenTextSplitter.
        :param chunk_overlap: The chunk overlap to use
            for the LlamaIndex TokenTextSplitter.
        :param stream: Whether to stream the chatbot or not.
        :param use_cache: Whether to use the cache or not.
        """

        chat = ChatOpenAI(
            model_name=model_name,
            temperature=temperature,
            streaming=True,
            verbose=True,
            callback_manager=BaseCallbackManager(
                handlers=[StreamingStdOutCallbackHandler()] if stream else []
            ),
        )
        llm_predictor = LLMPredictor(llm=chat)
        service_context = ServiceContext.from_defaults(llm_predictor=llm_predictor)

        # Initialize vector index or load it from disk.
        vector_index = GPTVectorStoreIndex(nodes=[], service_context=service_context)
        if saved_index_path is not None:
            vector_index = GPTVectorStoreIndex.load_from_disk(
                saved_index_path, service_context=service_context
            )

        # Override vector_index if doc_paths is specified.
        if doc_paths is not None:
            vector_index = make_or_load_vector_index(
                doc_paths, chunk_size, chunk_overlap, use_cache
            )

        # Set object attributes.
        self.system_message = system_message
        self.vector_index = vector_index
        self.doc_paths = doc_paths
        self.chat = chat
        self.chat_history = [
            SystemMessage(content=system_message),
            SystemMessage(
                content="""Do not hallucinate content.
If you cannot answer something, respond by saying that you don't know.
"""
            ),
        ]
        # Store a mapping of query to source nodes.
        self.source_nodes: dict = {}
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap

    # @validate_call
    def __call__(
        self,
        query: str,
        similarity_top_k: int = SIMILARITY_TOP_K,
    ) -> AIMessage:
        """Call the QueryBot.

        :param query: The query to send to the document index.
        :param similarity_top_k: The number of documents to return from the index.
            These documents are added to the context of the chat history
            and then used to synthesize the response.
            Default value is dynamically set based on the default model.
        :raises ValueError: if the index is not initialized.
        :return: The response to the query generated by GPT4.
        """
        if self.vector_index is None:
            raise ValueError(
                "You need to provide a document for querybot to index against!"
            )
        # Step 1: Get documents from the index that are deemed to be matching the query.
        # logger.info(f"Querying index for top {similarity_top_k} documents...")
        retriever = VectorIndexRetriever(
            index=self.vector_index, similarity_top_k=similarity_top_k
        )
        source_nodes = retriever.retrieve(query)
        source_texts = [n.node.text for n in source_nodes]

        faux_chat_history = []
        faux_chat_history.append(SystemMessage(content=self.system_message))

        # Step 2: Grab the last four responses from the chat history.
        faux_chat_history.extend(self.chat_history[-4:])

        # Step 2: Construct a faux message history to work with.
        faux_chat_history.append(
            SystemMessage(content="Here is the context you will be working with:")
        )
        for text in source_texts:
            faux_chat_history.append(SystemMessage(content=text))

        faux_chat_history.append(
            SystemMessage(content="Based on this context, answer the following query:")
        )

        faux_chat_history.append(HumanMessage(content=query))

        # Step 3: Send the chat history through the model
        response = self.chat(faux_chat_history)

        # Step 4: Record only the human response
        # and the GPT response but not the original.
        self.chat_history.append(HumanMessage(content=query))
        self.chat_history.append(response)

        # Step 5: Record the source nodes of the query.
        self.source_nodes[query] = source_nodes

        autorecord(query, response.content)

        # Step 6: Return the response.
        return response

    # @validate_call
    def save(self, path: Union[str, Path]):
        """Save the QueryBot index to disk.

        :param path: The path to save the QueryBot index.
        """
        path = Path(path)
        if not path.suffix == ".json":
            path = path.with_suffix(".json")
        self.vector_index.save_to_disk(path)

    def retrieve(
        self,
        query: str,
        similarity_top_k: int = SIMILARITY_TOP_K,
    ):
        """Retrieve the source nodes associated with a query using similarity search.

        :param query: The query to send to the document index.
        :param similarity_top_k: The number of documents to return from the index.
        :raises ValueError: if the index is not initialized.
        :return: The source nodes associated with the query.
        """
        if self.vector_index is None:
            raise ValueError(
                "You need to provide a document for querybot to index against!"
            )
        retriever = VectorIndexRetriever(
            index=self.vector_index, similarity_top_k=similarity_top_k
        )
        source_nodes = retriever.retrieve(query)
        return source_nodes


def make_service_context():
    """Make a service context for the QueryBot.

    :returns: A service context.
    """
    chat = ChatOpenAI(
        model_name=default_language_model(),
        temperature=0.0,
        streaming=True,
        verbose=True,
        callback_manager=BaseCallbackManager([StreamingStdOutCallbackHandler()]),
    )
    llm_predictor = LLMPredictor(llm=chat)
    service_context = ServiceContext.from_defaults(llm_predictor=llm_predictor)
    return service_context


# @validate_call
def load_index(persist_dir: Path, service_context: ServiceContext):
    """Load an index from disk.

    :param persist_dir: The directory to load the index from.
    :param service_context: The service context to use for the index.
    :returns: The index.
    """
    storage_context = StorageContext.from_defaults(
        docstore=SimpleDocumentStore.from_persist_dir(persist_dir=persist_dir),
        vector_store=SimpleVectorStore.from_persist_dir(persist_dir=persist_dir),
        index_store=SimpleIndexStore.from_persist_dir(persist_dir=persist_dir),
    )
    index = load_index_from_storage(storage_context, service_context=service_context)
    return index


def make_default_storage_context() -> StorageContext:
    """Make a default storage context for the QueryBot.

    :returns: A storage context.
    """
    storage_context = StorageContext.from_defaults(
        docstore=SimpleDocumentStore(),
        vector_store=SimpleVectorStore(),
        index_store=SimpleIndexStore(),
    )
    return storage_context


def make_vector_index(
    docs: List[Document],
    file_hash: str,
    persist_dir: Path,
    service_context: ServiceContext,
    chunk_size: int = 2000,
    chunk_overlap: int = 0,
) -> GPTVectorStoreIndex:
    """Make an index from a list of documents.

    :param docs: A list of documents to index.
    :param file_hash: The hash of the file to use as the index id.
    :param persist_dir: The directory to persist the index to.
    :param service_context: The service context to use for the index.
    :param chunk_size: The chunk size to use for the llama_index text splitter.
    :param chunk_overlap: The chunk overlap to use for the llama_index text splitter.
    :returns: The index.
    """
    # create parser and parse document into nodes
    parser = SimpleNodeParser.from_defaults(
        chunk_size=chunk_size, chunk_overlap=chunk_overlap
    )
    nodes = parser.get_nodes_from_documents(docs)

    # create (or load) docstore and add nodes
    storage_context = make_default_storage_context()
    storage_context.docstore.add_documents(nodes)

    vector_index = GPTVectorStoreIndex(
        nodes,
        storage_context=storage_context,
        index_id=file_hash,
        service_context=service_context,
    )
    vector_index.storage_context.persist(persist_dir=persist_dir)
    return vector_index


# @validate_call
def make_or_load_vector_index(
    doc_paths: List[Path] | List[str],
    chunk_size: int = 2000,
    chunk_overlap: int = 0,
    use_cache: bool = True,
) -> GPTVectorStoreIndex:
    """Make or load an index for a collection of documents.

    :param doc_paths: The path to the document to make or load an index for.
    :param chunk_size: The chunk size to use for the LangChain TokenTextSplitter.
    :param chunk_overlap: The chunk overlap to use for the LangChain TokenTextSplitter.
    :param use_cache: Whether to use the cache.
    :returns: The index.
    """
    # An index is constructed over a collection of documents.
    # As such, we construct a mapping of a hash of
    # the collection of documents to the index.

    # Step 1: Compute file hash for all of the documents
    # + the chunk size and chunk overlap.
    file_hash = hashlib.sha256()
    from tqdm.auto import tqdm

    for doc_path in tqdm(doc_paths, desc="hash files"):
        file_hash.update(Path(doc_path).read_bytes())
    file_hash.update(str(chunk_size).encode())
    file_hash.update(str(chunk_overlap).encode())
    file_hash_hexdigest = file_hash.hexdigest()

    # Make persist_dir based on the file hash's hexdigest.
    persist_dir = CACHE_DIR / file_hash_hexdigest
    persist_dir.mkdir(parents=True, exist_ok=True)
    service_context = make_service_context()

    # Step 2: Create the index's split documents.
    split_docs = []
    for doc_path in tqdm(doc_paths, desc="split documents"):
        document = magic_load_doc(doc_path)
        splitted_document = split_document(document[0], chunk_size, chunk_overlap)
        split_docs.extend(splitted_document)

    # Check that the persist directory exists and that we have made a docstore,
    # which is a sentinel test for the rest of the index.
    if persist_dir.exists() and (persist_dir / "docstore.json").exists() and use_cache:
        return load_index(persist_dir, service_context=service_context)

    return make_vector_index(
        split_docs,
        file_hash_hexdigest,
        persist_dir,
        service_context=service_context,
        chunk_size=chunk_size,
        chunk_overlap=chunk_overlap,
    )



"""Code for document preprocessing."""


from pathlib import Path
from typing import List

from langchain.text_splitter import TokenTextSplitter
from llama_index import Document, download_loader

EXTENSION_LOADER_MAPPING = {
    ".pdf": "PDFReader",
    ".docx": "DocxReader",
    ".pptx": "PptxReader",
    ".xlsx": "PandasExcelReader",
    ".md": "MarkdownReader",
    ".ipynb": "IPYNBReader",
}


def magic_load_doc(file_path: Path) -> List[Document]:
    """Load a document from a file.

    This function is used to magically load a document from a file.
    The file extension is inferred from the file path.
    If the file extension is not recognized, it is assumed to be a plain text file.

    :param file_path: Path to the file to be loaded.
    :return: A list of documents.
    """
    loader_string: str = EXTENSION_LOADER_MAPPING.get(Path(file_path).suffix, None)
    if loader_string is not None:
        # Treat this as a document that needs special processing.
        Loader = download_loader(loader_string)
        loader = Loader()
        documents = loader.load_data(file_path)

        # Concatenate the documents if there are more than 1 document in documents:
        if len(documents) > 1:
            documents = [Document(text=" ".join([doc.text for doc in documents]))]

    else:
        # Treat this as a plain text file.
        with open(file_path, "r+") as f:
            documents = [Document(text=f.read())]
    return documents


def split_document(
    doc: Document,
    chunk_size: int = 2000,
    chunk_overlap: int = 0,
) -> List[Document]:
    """Split a document into sub-documents using token text splitter.

    This function is used to split a document
    into sub-documents using token text splitter,
    such that each sub-document has a maximum length of `chunk_size` tokens
    with `chunk_overlap` tokens overlap.

    :param doc: Document to be split.
    :param chunk_size: Maximum length of each sub-document.
    :param chunk_overlap: Number of tokens to overlap between each sub-document.
    :raises ValueError: If `chunk_overlap` is negative.
    :return: A list of sub-documents.
    """
    if chunk_overlap < 0:
        raise ValueError("chunk_overlap must be non-negative.")
    splitter = TokenTextSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)
    sub_texts = splitter.split_text(doc.text)

    sub_docs = [Document(text=t) for t in sub_texts]
    return sub_docs



"""
This module contains tests for the llamabot.file_finder module.

It tests the following functions:
- recursive_find
- check_in_git_repo
- read_file

It uses the following external libraries:
- pytest
- hypothesis
"""
import os
import subprocess
import tempfile
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from hypothesis import given
from hypothesis import strategies as st

from llamabot.file_finder import check_in_git_repo, read_file, recursive_find


@given(st.text(min_size=1, max_size=5, alphabet="abcdefghijklmnopqrstuvwxyz"))
def test_recursive_find_empty_directory(extension):
    """Test that recursive_find returns an empty list when given an empty directory.

    :param extension: The extension to search for.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        root = Path(temp_dir)
        assert recursive_find(root, extension) == []


@given(
    st.text(min_size=1, max_size=5, alphabet="abcdefghijklmnopqrstuvwxyz"),
    st.text(min_size=1, max_size=5, alphabet="abcdefghijklmnopqrstuvwxyz"),
)
def test_recursive_find_single_file(extension, file_name):
    """Test that recursive_find returns a single file when given a directory with a single file.

    :param extension: The extension to search for.
    :param file_name: The name of the file to create.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        root = Path(temp_dir)
        file_path = root / f"{file_name}.{extension}"
        file_path.touch()
        assert recursive_find(root, f".{extension}") == [file_path]


@given(
    st.text(min_size=1, max_size=5, alphabet="abcdefghijklmnopqrstuvwxyz"),
    st.text(min_size=1, max_size=5, alphabet="abcdefghijklmnopqrstuvwxyz"),
)
def test_recursive_find_nested_file(extension, file_name):
    """Test that recursive_find returns a single file when given a directory with a single file.

    :param extension: The extension to search for.
    :param file_name: The name of the file to create.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        root = Path(temp_dir)
        nested_dir = root / "nested"
        nested_dir.mkdir()
        file_path = nested_dir / f"{file_name}.{extension}"
        file_path.touch()
        assert recursive_find(root, f".{extension}") == [file_path]


@given(
    st.text(min_size=1, max_size=5, alphabet="abcdefghijklmnopqrstuvwxyz"),
    st.text(min_size=1, max_size=5, alphabet="abcdefghijklmnopqrstuvwxyz"),
)
def test_recursive_find_multiple_files(extension, file_name):
    """Test that recursive_find returns multiple files when given a directory with multiple files.

    :param extension: The extension to search for.
    :param file_name: The name of the file to create.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        root = Path(temp_dir)
        file_path1 = root / f"{file_name}.{extension}"
        file_path1.touch()
        nested_dir = root / "nested"
        nested_dir.mkdir()
        file_path2 = nested_dir / f"{file_name}.{extension}"
        file_path2.touch()
        assert set(recursive_find(root, f".{extension}")) == {file_path1, file_path2}


def test_check_in_git_repo_true():
    """
    Test check_in_git_repo returns True when git command returns 'true'.
    """
    subprocess.check_output = MagicMock(return_value=b"true\n")
    assert check_in_git_repo("some_path") is True


def test_check_in_git_repo_false():
    """
    Test check_in_git_repo returns False when git command returns 'false'.
    """
    subprocess.check_output = MagicMock(return_value=b"false\n")
    assert check_in_git_repo("some_path") is False


def test_check_in_git_repo_error():
    """
    Test check_in_git_repo raises an exception when git command raises an exception.
    """
    subprocess.check_output = MagicMock(
        side_effect=subprocess.CalledProcessError(1, "git")
    )
    with pytest.raises(subprocess.CalledProcessError):
        check_in_git_repo("some_path")


@given(st.text(alphabet="abcdefghijklmnopqrstuvwxyz"))
def test_read_file_contents(text: str):
    """
    Test if read_file returns the correct contents of a file.

    :param text: The text to write to the file.
    """
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        temp_file.write(text.encode())
        temp_file_path = Path(temp_file.name)

    assert read_file(temp_file_path) == text
    os.remove(temp_file_path)


def test_read_file_nonexistent_path():
    """
    Test if read_file raises a FileNotFoundError for a nonexistent file path.
    """
    non_existent_path = Path("non_existent_file.txt")

    with pytest.raises(FileNotFoundError):
        read_file(non_existent_path)



"""Prompts and bots for coding.

This module provides a codebot instance and several functions for generating code, docstrings, and tests.
It also includes a function to create a panel app for codebot.

Functions:
    - ghostwriter(desired_functionality, language)
    - docstring(code)
    - module_doc(source_file)
    - tests(code, source_file)
    - create_panel_app()
"""
import panel as pn

from llamabot.bot.simplebot import SimpleBot
from llamabot.panel_utils import PanelMarkdownCallbackHandler

try:
    from outlines import text
except ImportError:
    import warnings

    warnings.warn(
        "Please install the `outlines` package to use the llamabot prompt library."
    )


def codebot() -> SimpleBot:
    """Return a codebot instance.

    :return: A codebot instance.
    """
    return SimpleBot(
        """You are a programming expert.

    You provide suggestions to programmers in Python by default,
    but can also suggest in other programming languages when prompted.

    Please write code without explaining it.
    Do not explain your code, only provide code.

    In your code, prefer the use of the Pathlib module over the os module for handling paths.
    """
    )


@text.prompt
def ghostwriter(desired_functionality, language):
    """I would like to accomplish the following.

    {{ desired_functionality }}

    Please return code in the {{ language }} programming language.

    If writing in Python, ensure that there are type hints in the function.
    Ensure that the implementation of the function results in
    the simplest type hints possible.

    Ensure that within any errors raised, the error message is actionable
    and informs the user what they need to do to fix the error.
    Make sure the error message is prescriptive,
    possibly even more verbose than necessary,
    and includes verbiage such as "please do X" or "please do not do Y".
    # noqa: DAR101
    """


@text.prompt
def docstring(code, style="sphinx"):
    """Please help me write docstrings for the following code.

    {{ code }}

    Ensure that the docstring is written in {{ style }} style.

    Ensure that the code usage example is located before the arguments documentation.
    Ensure that the code usage example is renderable using Markdown directives.

    Do not include any typing information in the docstring
    as they should be covered by the type hints.

    Do not include the original source code in your output,
    return only the docstring starting from the triple quotes
    and ending at the triple quotes.
    Do not include the original function signature either.
    Write only the docstring and nothing else.
    Ensure that the docstring indentation is correct for the object.

    # noqa: DAR101
    """


@text.prompt
def module_doc(source_file_contents, source_file_fpath=None, file_tree=None):
    """Please help me write module-level docstrings for the following code.

    {{ source_file_contents }}

    For context, this is the source file's path:

    {{ source_file_fpath }}

    And this is the file tree of the source file's directory:

    {{ file_tree }}

    Module-level docstrings have the following specification:

    The docstring for a module should generally list the classes,
    exceptions and functions (and any other objects)
    that are exported by the module, with a one-line summary of each.
    (These summaries generally give less detail
    than the summary line in the object's docstring.)
    The docstring for a package
    (i.e., the docstring of the package's __init__.py module)
    should also list the modules and subpackages exported by the package.

    Ensure that you never spit out the original source code.

    # noqa: DAR101
    """


@text.prompt
def tests(code, source_file_txt: str, dependent_source_files: str):
    """I need help writing unit tests.

    For context, here is the broader source file in which the code is defined:

    {{ source_file_txt }}

    Then, there are other source files that this function depends on,
    from which you can import objects.

    {{ dependent_source_files }}

    Prefer the use of property-based tests over example-based tests.
    Only suggest example-based tests
    if it is too difficult to generate property-based tests.
    For each test function,
    ensure that there is a docstring that explains what the test is testing.

    Use pytest-style test functions and not Unittest-style test classes.

    Here is the actual thing for which I need a test:

    {{ code }}

    Please write me tests for that code.

    # noqa: DAR101
    """


def create_panel_app() -> pn.Column:
    """Create a panel app for codebot.

    :return: The panel app.
    """
    programming_languages = [
        "Python",
        "Java",
        "JavaScript",
        "C",
        "C++",
        "C#",
        "Ruby",
        "Go",
        "Rust",
        "Swift",
        "Kotlin",
        "TypeScript",
        "PHP",
        "Perl",
        "Objective-C",
        "Shell",
        "SQL",
        "HTML",
        "CSS",
        "R",
        "MATLAB",
        "Scala",
        "Groovy",
        "Lua",
        "Haskell",
        "Elixir",
        "Julia",
        "Dart",
        "VB.NET",
        "Assembly",
        "F#",
    ]
    language = pn.widgets.Select(name="Select Language", options=programming_languages)

    user_specification = pn.widgets.TextAreaInput(
        name="User Specification",
        placeholder="Please write me a function that generates Fibonacci numbers.",
    )
    code_output = pn.pane.Markdown()
    code_output.object = "_Generated code will show up here._"
    test_output = pn.pane.Markdown()
    test_output.object = "_Generated unit tests will show up here._"

    def generate_code(event):
        """Callback for the code generator button.

        :param event: The button click event.
        """
        code_output.object = f"```{language.value}\n"
        markdown_handler = PanelMarkdownCallbackHandler(code_output)
        codebot.model.callbacks.set_handler(markdown_handler)
        code_text = codebot(ghostwriter(user_specification.value, language.value))
        code_output.object = f"```{language.value}\n{code_text.content}\n```"

    def generate_tests(event):
        """Callback for the test generator button.

        :param event: The button click event.
        """
        test_output.object = f"```{language.value}\n"
        markdown_handler = PanelMarkdownCallbackHandler(test_output)
        codebot.model.callbacks.set_handler(markdown_handler)
        test_text = codebot(tests(code_output.object, language.value))
        test_output.object = f"```{language.value}\n{test_text.content}\n```"

    generate_button = pn.widgets.Button(name="Generate Code")
    generate_button.on_click(generate_code)

    test_button = pn.widgets.Button(name="Generate Unit Tests")
    test_button.on_click(generate_tests)

    return pn.Column(
        language,
        user_specification,
        generate_button,
        code_output,
        test_button,
        test_output,
    )



"""Zotero prompt library."""
try:
    from outlines import text
except ImportError:
    import warnings

    warnings.warn(
        "Please install the `outlines` package to use the llamabot prompt library."
    )


@text.prompt
def retrieverbot_sysprompt():
    """You are an expert in retrieving information from JSON files."""


@text.prompt
def docbot_sysprompt():
    """You are an expert in answering questions about any paper.

    Your responses are like how Richard Feynman would answer questions.
    That is to say, you will explain yourself and the paper (where relevant)
    at a level that an undergraduate student would understand.
    """


@text.prompt
def get_key(query: str = ""):
    """Based on the context provided,
    return for me all of the keys that were provided for the query.

    Ensure that you're using the titles and abstracts.

    {{ query }}

    Return this as JSON:

    {
        "key": [<key1>, <key2>, ...]
    }

    The list should include only the keys for the papers that are relevant to the query.

    Do not return anything else except JSON.
    Ensure that you do not return any explanations of your response.
    If you cannot help with the request, return an empty list.

    # noqa: DAR101
    """



"""Zotero utilities."""
import os

from pyzotero import zotero


def load_zotero():
    """Load Zotero API client.

    :return: Zotero API client
    """
    zotero_library_id = os.environ.get("ZOTERO_LIBRARY_ID", None)
    zotero_library_type = os.environ.get("ZOTERO_LIBRARY_TYPE", None)
    zotero_api_key = os.environ.get("ZOTERO_API_KEY", None)

    return zotero.Zotero(
        library_id=zotero_library_id,
        library_type=zotero_library_type,
        api_key=zotero_api_key,
    )



"""Bot abstractions that let me quickly build new GPT-based applications."""

import os
import warnings

import openai
import panel as pn
from dotenv import load_dotenv

from llamabot.config import llamabotrc_path

from .chatbot import ChatBot
from .querybot import QueryBot
from .simplebot import SimpleBot

pn.extension()
load_dotenv()

if llamabotrc_path.exists():
    load_dotenv(llamabotrc_path)

api_key = os.getenv("OPENAI_API_KEY", None)
if api_key is None:
    warnings.warn(
        "No OpenAI API key found. Please set OPENAI_API_KEY in your environment.",
        category=RuntimeWarning,
    )
openai.api_key = api_key


__all__ = ["SimpleBot", "ChatBot", "QueryBot"]



"""Tests for the code_manipulation module."""

import os
import tempfile
from pathlib import Path

import pytest

from llamabot.code_manipulation import (
    get_dependencies,
    get_object_source_code,
    insert_docstring,
    replace_object_in_file,
)


def test_insert_docstring_function():
    """Test that a docstring is inserted into a function successfully.

    This test tests that a docstring is inserted into a function successfully.
    """
    source_code = "def foo():\n    pass"
    object_name = "foo"
    new_docstring = "This is a test docstring."

    with tempfile.NamedTemporaryFile("w+") as temp_file:
        temp_file.write(source_code)
        temp_file.flush()

        insert_docstring(temp_file.name, object_name, new_docstring)

        with open(temp_file.name, "r") as updated_file:
            updated_code = updated_file.read()

    expected_code = f'def {object_name}():\n    """{new_docstring}"""\n    pass\n'
    assert updated_code == expected_code


def test_insert_docstring_class():
    """Test that a docstring is inserted into a class successfully."""
    source_code = "class Foo:\n    pass"
    object_name = "Foo"
    new_docstring = "This is a test docstring."

    with tempfile.NamedTemporaryFile("w+") as temp_file:
        temp_file.write(source_code)
        temp_file.flush()

        insert_docstring(temp_file.name, object_name, new_docstring)

        with open(temp_file.name, "r") as updated_file:
            updated_code = updated_file.read()

    expected_code = f'class {object_name}:\n    """{new_docstring}"""\n    pass\n'
    assert updated_code == expected_code


@pytest.mark.xfail(
    reason="18 June 2023: Fails because of quotation marks. Need to fix."
)
def test_get_object_source_code_function():
    """Test that the source code of a function is returned successfully."""
    source_code = """
def test_function():
    return "Hello, World!"
    """
    with open("test_file.py", "w") as f:
        f.write(source_code)

    result = get_object_source_code("test_file.py", "test_function")
    assert result.strip() == source_code.strip()

    Path("test_file.py").unlink()


@pytest.mark.xfail(
    reason="18 June 2023: Fails because of quotation marks. Need to fix."
)
def test_get_object_source_code_class():
    """Test that the source code of a class is returned successfully."""
    source_code = """
class TestClass:
    def __init__(self):
        self.value = "Hello, World!"
    """
    with open("test_file.py", "w") as f:
        f.write(source_code)

    result = get_object_source_code("test_file.py", "TestClass")
    assert result.strip() == source_code.strip()

    Path("test_file.py").unlink()


def test_get_object_source_code_not_found():
    """Test that a NameError is raised when the object is not found."""
    source_code = """
def test_function():
    return "Hello, World!"
    """
    with open("test_file.py", "w") as f:
        f.write(source_code)

    with pytest.raises(NameError):
        get_object_source_code("test_file.py", "non_existent_function")

    Path("test_file.py").unlink()


def test_get_object_source_code_syntax_error():
    """Test that a SyntaxError is raised when the source code has invalid Python syntax."""
    source_code = """
def test_function()
    return "Hello, World!"
    """
    with open("test_file.py", "w") as f:
        f.write(source_code)

    with pytest.raises(SyntaxError):
        get_object_source_code("test_file.py", "test_function")

    Path("test_file.py").unlink()


def test_replace_function_in_file():
    """
    Test that replace_object_in_file correctly replaces a function in a source file.
    """
    source_code = """
def foo():
    return 42
"""
    new_function_definition = """
def foo():
    return 0
"""

    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as temp_file:
        temp_file.write(source_code)
        temp_file.flush()

        replace_object_in_file(temp_file.name, "foo", new_function_definition)

        with open(temp_file.name, "r") as updated_file:
            updated_code = updated_file.read()

    assert updated_code.strip() == new_function_definition.strip()


def test_replace_class_in_file():
    """
    Test that replace_object_in_file correctly replaces a class in a source file.
    """
    source_code = """
class Foo:
    pass
"""
    new_class_definition = """
class Foo:

    def __init__(self):
        self.x = 42
"""

    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as temp_file:
        temp_file.write(source_code)
        temp_file.flush()

        replace_object_in_file(temp_file.name, "Foo", new_class_definition)

        with open(temp_file.name, "r") as updated_file:
            updated_code = updated_file.read()

    assert updated_code.strip() == new_class_definition.strip()


def test_replace_object_in_file_syntax_error():
    """
    Test that replace_object_in_file raises a SyntaxError when the source file has invalid Python syntax.
    """
    source_code = """
def foo():
    return 42
invalid_syntax
"""

    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as temp_file:
        temp_file.write(source_code)
        temp_file.flush()

        try:
            replace_object_in_file(temp_file.name, "foo", "def foo():\\n    return 0")
        except SyntaxError:
            pass
        else:
            assert False, "Expected a SyntaxError to be raised."


def test_replace_object_in_file_value_error():
    """
    Test that replace_object_in_file raises a ValueError when the specified object does not exist in the source file.
    """
    source_code = """
def foo():
    return 42
"""

    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as temp_file:
        temp_file.write(source_code)
        temp_file.flush()

        try:
            replace_object_in_file(temp_file.name, "bar", "def bar():\\n    return 0")
        except ValueError:
            pass
        else:
            assert False, "Expected a ValueError to be raised."


def test_get_dependencies():
    """Test the get_dependencies function.

    This test tests that the get_dependencies function returns a list of dependencies.
    """
    # Create a temporary directory
    with tempfile.TemporaryDirectory() as temp_dir:
        # Create a temporary source file
        source_code = """
import module1
from module2 import function2
from module3 import Class3

def function1():
    pass
        """
        source_file = os.path.join(temp_dir, "source_file.py")
        with open(source_file, "w") as file:
            file.write(source_code)

        # Create temporary dependency files
        module1_code = """
def function_module1():
    pass
        """
        module1_file = os.path.join(temp_dir, "module1.py")
        with open(module1_file, "w") as file:
            file.write(module1_code)

        module2_code = """
def function2():
    pass
        """
        module2_file = os.path.join(temp_dir, "module2.py")
        with open(module2_file, "w") as file:
            file.write(module2_code)

        module3_code = """
class Class3:
    pass
        """
        module3_file = os.path.join(temp_dir, "module3.py")
        with open(module3_file, "w") as file:
            file.write(module3_code)

        # Test get_dependencies function
        object_name = "function1"
        dependencies = get_dependencies(source_file, object_name)

        assert isinstance(dependencies, list)
        for dependency in dependencies:
            assert isinstance(dependency, str)
            assert os.path.isfile(dependency)
            assert os.path.splitext(dependency)[1] == ".py"

        # Additional assertions based on the specific dependencies
        assert module1_file in dependencies
        assert module2_file in dependencies
        assert module3_file in dependencies

        # Clean up temporary files and directory (optional)
        # You can comment out the following lines if you want to inspect the temporary files manually
        os.remove(source_file)
        os.remove(module1_file)
        os.remove(module2_file)
        os.remove(module3_file)



"""LlamaBot configuration."""
import openai
import typer
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter
import os
from .utils import configure_environment_variable

app = typer.Typer()


@app.command()
def api_key(
    api_key: str = typer.Option(
        ..., prompt=True, hide_input=True, confirmation_prompt=True
    ),
) -> None:
    """
    Configure the API key for llamabot.

    .. code-block:: python

        configure(api_key="your_api_key_here")

    :param api_key: The API key to be used for authentication.
    """
    configure_environment_variable(env_var="OPENAI_API_KEY", env_value=api_key)


@app.command()
def default_model(model_name=None):
    """Configure the default model for llamabot.

    If no model name is provided,
    or if the model name is not one of those that the user's API key has access to,
    then the user will be prompted for a model name.

    :param model_name: The name of the model to be used for default
    """
    from dotenv import load_dotenv

    from llamabot.config import llamabotrc_path

    load_dotenv(llamabotrc_path)

    openai.api_key = os.getenv("OPENAI_API_KEY")

    model_list = openai.Model.list()["data"]
    available_models = [x["id"] for x in model_list if "gpt" in x["id"]]
    available_models.sort()

    if model_name in available_models:
        configure_environment_variable(
            env_var="OPENAI_DEFAULT_MODEL", env_value=model_name
        )
        return

    completer = WordCompleter(available_models)

    typer.echo("These are the GPT models available to you:")
    for model in available_models:
        typer.echo(model)

    while True:
        default_model = prompt(
            "Please type the name of the model you'd like to use: ",
            completer=completer,
            complete_while_typing=True,
            default=available_models[-1],
        )
        if default_model in available_models:
            configure_environment_variable(
                env_var="OPENAI_DEFAULT_MODEL", env_value=default_model
            )
            break



"""Prompts for writing git stuff."""
from llamabot import SimpleBot
import os

try:
    from outlines import text
except ImportError:
    import warnings

    warnings.warn(
        "Please install the `outlines` package to use the llamabot prompt library."
    )


def commitbot():
    """Return a commitbot instance.

    It is hard-coded to use gpt-3.5-turbo-16k-0613 as its model.
    This model is sufficient for the quality of commit messages,
    matching gpt-4-32k but being 1/10 the cost.

    :return: A commitbot instance.
    """
    return SimpleBot(
        "You are an expert user of Git.",
        model_name=os.getenv("OPENAI_DEFAULT_MODEL", "gpt-3.5-turbo-16k-0613"),
    )


@text.prompt
def write_commit_message(diff: str):
    """Please write a commit message for the following diff.

    {{ diff }}

    # noqa: DAR101

    Use the Conventional Commits specification to write the diff.

    [COMMIT MESSAGE BEGIN]
    <type>[optional scope]: <description>

    [optional body]

    [optional footer(s)]
    [COMMIT MESSAGE END]

    The commit contains the following structural elements,
    to communicate intent to the consumers of your library:

    fix: a commit of the type fix patches a bug in your codebase
        (this correlates with PATCH in Semantic Versioning).
    feat: a commit of the type feat introduces a new feature to the codebase
        (this correlates with MINOR in Semantic Versioning).
    BREAKING CHANGE: a commit that has a footer BREAKING CHANGE:,
        or appends a ! after the type/scope,
        introduces a breaking API change
        (correlating with MAJOR in Semantic Versioning).
        A BREAKING CHANGE can be part of commits of any type.

    types other than fix: and feat: are allowed,
    for example @commitlint/config-conventional
    (based on the Angular convention) recommends
    build:, chore:, ci:, docs:, style:, refactor:, perf:, test:, and others.

    footers other than BREAKING CHANGE: <description> may be provided
    and follow a convention similar to git trailer format.

    Additional types are not mandated by the Conventional Commits specification,
    and have no implicit effect in Semantic Versioning
    (unless they include a BREAKING CHANGE).
    A scope may be provided to a commit's type,
    to provide additional contextual information and is contained within parenthesis,
    e.g., feat(parser): add ability to parse arrays.
    Within the optional body section, prefer the use of bullet points.

    Final instructions:

    1. Do not fence the commit message with back-ticks or quotation marks.
    2. Do not add any other text except the commit message itself.
    3. Only write out the commit message.

    [BEGIN COMMIT MESSAGE]
    """


@text.prompt
def compose_release_notes(commit_log, repo_url):
    """Here is a commit log:

    # noqa: DAR101

    {{ commit_log }}

    These come from the following repo:

    {{ repo_url }}

    Please write for me the release notes.
    The notes should contain a human-readable summary
    of each new feature that was added.

    Follow the following format:

        ## Version <version number>

        <brief summary of the new version>

        ### New Features

        - <describe in plain English> (<commit's first 6 letters>) (<commit author>)
        - <describe in plain English> (<commit's first 6 letters>) (<commit author>)

        ### Bug Fixes

        - <describe in plain English> (<commit's first 6 letters>) (<commit author>)

        ### Deprecations

        - <describe in plain English> (<commit's first 6 letters>) (<commit author>)
    """
